//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//						(c) 1997 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	spec.cc -- implementation of the Spec class
//
//	CHANGE HISTORY:
//              DAH, 25/03/98: 1.0.0: - first version
//              DAH, 06/04/98: 1.0.1: - make _copyGlobalAttribs method
//				UL, 2000-12-20: 2.0.0: - couple of trivial changes to make the
//										 optimizer happy
//
//	AUTHOR:
//              D. Hinshaw
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Expanded description of SAS task `evselect'
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/task_devel/packages/evselect/src/spec.cc,v 1.34 2000/12/22 13:36:12 ulammers Exp $

#include <iostream>
#include <cmath>
#include <cstdio>

#include "param.h"
#include "Dal.h"
#include "errstr.h"
#include <ExprColumnGenerator.h>
#include "Extractor.h"
#include "DataSubspace.h"
#include "DataFilter.h"
#include "xmmexposure.h"

void Spec::make(bool writeDss){

  // Read in the needed parameters

  int IbinSize = intParameter("spectralbinsize");
  bool zcol = booleanParameter("withzcolumn");
  bool zerrcol = booleanParameter("withzerrorcolumn");

  // Get pointers to the needed columns

  Column *eDatCol, *zDatCol = 0, *zDatErrCol = 0;
  Column *eCol, *zCol, *zErrCol;

  std::string ecolumn = stringParameter("energycolumn");
  if( _intable->hasColumn(ecolumn) ){
    eCol = _intable->column(ecolumn);

    if( eCol->dataType()!=Dal::Int8 && eCol->dataType()!=Dal::Int16 && 
	eCol->dataType()!=Dal::Int32 )
      errstr << "Energy column is not integer-valued.  All values will be "
             << "truncated to integers before accumulation of spectrum."
	     << Sas::warning("SpecIntColumn");

    eDatCol = reType(eCol, Dal::Int32);
  }else{
    std::strstream newName;
    newName << "E" << _intable << ends;
    ExprColumnGenerator colGen(ecolumn, newName.str(), Dal::Int32);
    colGen.doIt(*_intable, *_tempTable);
    eCol = eDatCol = _intable->column(ecolumn);
  }

  if(zcol){
    if( !_intable->parent()->hasAttribute("TELESCOP") ||
	_intable->parent()->attribute("TELESCOP")->asString() != "XMM" ){
      errstr << "Not XMM data.  Cannot calculate a weighted spectrum for non-XMM data"
	     << Sas::warning("NotXmmSpectrum");
      return;
    }

    std::string zcolumn = stringParameter("zcolumn");
    if( _intable->hasColumn(zcolumn) ){
      zCol = _intable->column(zcolumn);
      zDatCol=reType(zCol, Dal::Real64);
    }else{
      std::strstream newName;
      newName << "Z" << _intable << ends;
      ExprColumnGenerator colGenZ(zcolumn, newName.str());
      colGenZ.doIt(*_intable, *_tempTable);
      zCol = zDatCol = _intable->column(zcolumn);
    }

    if(zerrcol){
      zcolumn = stringParameter("zerrorcolumn");
      if( _intable->hasColumn(zcolumn) ){
	zErrCol = _intable->column(zcolumn);
	zDatErrCol=reType(zErrCol, Dal::Real64);
      }else{
	std::strstream newName;
	newName << "ZE" << _intable << ends;
	ExprColumnGenerator colGenZ(zcolumn, newName.str());
	colGenZ.doIt(*_intable, *_tempTable);
	zErrCol = zDatErrCol = _intable->column(zcolumn);
      }
    }
  }

  // Determine the energy range

  std::vector<real64> limits(2);
  GetLimits(eCol, "specchannelmin", "specchannelmax", booleanParameter("withspecranges"), limits);

  std::vector<int32> useRange(2);
  useRange[0] = (int32)limits[0];
  useRange[1] = (int32)limits[1];
  std::vector<int32> tlRange(2);

  if( useRange[1] == useRange[0] )
    errstr << "The energy column for extracting the spectrum has "
	   << "the same value throughout, or the upper and lower limits "
	   << "for this column are equal. An empty spectrum will "
	   << "be created." << Sas::warning("BadSpecLimits"); 

  if( useRange[1] < useRange[0] )
    errstr << "Upper spectral limit < lower limit.  Cannot continue." 
	   << Sas::error("BadSpecLimits");

  unsigned long nRows = (unsigned long) ceil( ((float)(useRange[1]-useRange[0]+1))/IbinSize );

  // Now initialize the output spectrum

  Table * outTable = _dataSet->addTable("SPECTRUM", nRows);

  _copyGlobalAttribs(outTable);

  // Write some keywords

  *outTable->addAttribute( "HDUCLASS",
			   "Format conforms to OGIP/GSFC conventions")
    = "OGIP";
  *outTable->addAttribute( "HDUCLAS1",
			    "File contains a spectrum")
    = "SPECTRUM";
  
  *outTable->addAttribute( "HDUCLAS2",
			   "File contains gross counts")
    = "TOTAL";

  *outTable->addAttribute( "HDUCLAS3",
			   "Spectrum is stored as counts")
    = "COUNT";

  *outTable->addAttribute( "HDUVERS1",
			   "Version of format")
    = "1.1.0";

  *outTable->addAttribute( "AREASCAL",
			   "Nominal scaling factor for data")
    = 1.0;

  *outTable->addAttribute( "BACKSCAL",
			   "Scaling factor for background")
    = 1.0;

  *outTable->addAttribute( "CORRSCAL",
			   "Nominal scaling factor for correction file")
    = 1.0;

  *outTable->addAttribute( "POISSERR",
			   "Poisson errors appropriate")
    = true;

  *outTable->addAttribute( "SYS_ERR",
			   "Global systematic error")
    = 0;

  *outTable->addAttribute( "CHANTYPE","Type of channel data") = eCol->name();

  *outTable->addAttribute( "DETCHANS",
			   "Total number of detector channels available")
    = (int)nRows;

  *outTable->addAttribute( "QUALITY",
			   "All channels have good quality")
    = 0;

  *outTable->addAttribute( "GROUPING",
			   "No data grouping done")
    = 0;

  *outTable->addAttribute( "SPECDELT",
			   "Spectral channel size, ie the binning factor")
    = IbinSize;

  *outTable->addAttribute( "SPECPIX",
			   "The rebinned channel correspondsing to SPECVAL")
    = 0;

  *outTable->addAttribute( "SPECVAL",
			   "Original chan value at center of rebinned chan")
    = useRange[0] - 0.5 + float(IbinSize)/2;

  // Make the necessary columns.
  
  Column * chanCol = outTable->addColumn("CHANNEL", Dal::Int16, "");
  ColumnData * cChanDat = chanCol->data();
  int16 * chanDat = cChanDat->int16Data();

  *chanCol->addAttribute("TLMIN") = 0;

  *chanCol->addAttribute("TLMAX") = (int)nRows-1;

  real64 *rateDat=0, *rateErrDat=0;
  real64 *zData=0, *zErrData=0;
  ColumnData *cRateDat=0, *cErrRateDat=0, *cZDatCol=0, *cZDatErrCol=0, *cCntDat=0;
  int32 * cntDat=0;

  if( zcol ){
    cRateDat = outTable->addColumn("RATE", Dal::Real64, "", "count/s")->data();
    cErrRateDat = outTable->addColumn("STAT_ERR", Dal::Real64, "", "count/s")->data();
    rateDat = cRateDat->real64Data();
    rateErrDat = cErrRateDat->real64Data();
    for(unsigned i=0; i<nRows; i++) rateDat[i] = rateErrDat[i] = 0;

    cZDatCol = zDatCol->data();
    zData = cZDatCol->real64Data();
    if(zerrcol){
      cZDatErrCol = zDatErrCol->data();
      zErrData = cZDatErrCol->real64Data();
    }
  }else{
    cCntDat = outTable->addColumn("COUNTS", Dal::Int32, "", "count")->data();
    cntDat = cCntDat->int32Data();
    for(unsigned i=0; i<nRows; i++) cntDat[i]=0;
  }
  
  for(unsigned i=0; i<nRows; i++)
    chanDat[i] = i;

  delete cChanDat;

  // Set up type converted columns

  Column *flagDatCol;
  ColumnData *cFlagData = 0, *cEData = eDatCol->data();
  int32 index, *flagData = 0, *eData = cEData->int32Data();

  if(! _destruct){
    flagDatCol = reType(_intable->column(_flagCol), Dal::Int32);
    cFlagData = flagDatCol->data();
    flagData = cFlagData->int32Data();
  }

  // Accumlate the spectrum

  unsigned long mRows = _intable->rows();
 
  if(zcol){
    // Get the CCD info necessary for calculating rates
    std::vector<real64> exposures(18);
    for(int i=1; i<=17; i++){
      std::strstream attName;
      attName << "LIVETI";
      attName.width(2);
      attName.fill('0');
      attName << i << ends;
      if(_intable->hasAttribute(attName.str()))
	exposures[i] = _intable->attribute(attName.str())->asReal();
    }

    ColumnData * cCcdDat = _intable->column("CCDNR")->data();
    int8 * ccdDat = cCcdDat->int8Data();
    
    for(unsigned long irow=0; irow<mRows; irow++){
      if( ! _destruct && (( flagData[irow] & _flagBit) == 0) ) continue;

      index = (eData[irow]-useRange[0])/IbinSize;
      if( index >= 0 && index < (int32)nRows ){
	rateDat[index] += zData[irow]/exposures[ccdDat[irow]];
	if(zerrcol)
	  rateErrDat[index] += zErrData[irow]*zErrData[irow]/(exposures[ccdDat[irow]]*exposures[ccdDat[irow]]);
	else
	  rateErrDat[index] += zData[irow]*zData[irow]/(exposures[ccdDat[irow]]*exposures[ccdDat[irow]]);
      }
    }

    for(unsigned long irow=0; irow<outTable->rows(); irow++)
      rateErrDat[irow] = sqrt(rateErrDat[irow]);
    
    delete cRateDat;
    delete cErrRateDat;
    delete cZDatCol;
    if(zerrcol) delete cZDatErrCol;

  }else{
    for(unsigned long irow=0; irow<mRows; irow++){
      if( ! _destruct && (( flagData[irow] & _flagBit) == 0) ) continue;
      index = (eData[irow]-useRange[0])/IbinSize;
      if( index >= 0 && index < (int32)nRows ) cntDat[index]++;
    }
    delete cCntDat;
  }

  delete cEData;

  if(! _destruct) delete cFlagData;

  // Write data subspace info
  if (writeDss) {
     DataSubspace dss(_intable);
     outTable << dss;
  }

  // Calculate exposure keyword

  _getExposure(outTable);

  // Done
}
